// ignore_for_file: public_member_api_docs

import 'dart:developer';

import 'package:fl_chart/fl_chart.dart';
import 'package:healthrian_ble_support_for_wearcardiot/'
    'healthrian_ble_support_for_wearcardiot.dart';
import 'package:healthrian_common_support/healthrian_common_support.dart';
import 'package:lzma/lzma.dart';

mixin Ecg {
  /// [maxBufferStreamEcg] decides a maximum length of buffered Stream.
  ///
  /// For 40 seconds buffer and 250 sampling rate,
  /// It uses 40 * 250.
  int maxBufferStreamEcg = 40 * 250;

  /// [lengthBufferStreamEcg] decides a length of buffered Stream,
  /// after flushing buffer when reaching a maximum size.
  ///
  /// For 30 seconds buffer and 250 sampling rate,
  /// It uses 30 * 250.
  int lengthBufferStreamEcg = 30 * 250;

  /// [isFilteredEcg] decides whether ecg signal is filtered.
  bool isFilteredEcg = true;

  /// [isCompressedEcg] decides whether ecg signal is compressed.
  ///
  /// If it is `true`, ecg signal is decompressed by LZMA algorithm.
  bool isCompressedEcg = false;

  final _dataFilter = DataFilter(1);

  final _adaptiveHR = AdaptiveHR(0);

  bool isPaused = false;
}

extension BLEEcg on BLEManager {
  double _filter(double value) => _dataFilter.applyFilter(0, value, true);

  /// [startMeasurementEcg] sends read_ecg_packet to device,
  ///  and then, it registers all ecg streams of [streams].
  Future<void> startMeasurementEcg() async {
    await _sendCommandReadEcg();

    final isRunnable = await _checkValidation();
    if (!isRunnable) return;

    streams
      ..rawStreamEcg = _getStreamEcg().broadcast
      ..decompressedStreamEcg = _getDecompressedStreamEcg().broadcast
      ..recentStreamEcg = _getRecentStreamEcg().broadcast
      ..leadStateStreamEcg = _getLeadStateStream().broadcast
      ..heartRateStreamEcg = _getHeartRateStream().broadcast
      ..bufferedStreamEcg = _getBufferedStreamEcg().broadcast
      ..bufferedPlotStreamEcg = _getBufferedPlotStreamEcg().broadcast;

    isPaused = false;
  }

  /// [pauseMeasurementEcg] pause measurements for Ecg.
  Future<void> pauseMeasurementEcg() async {
    if (isPaused) {
      await startMeasurementEcg();
      return;
    }
    isPaused = true;
    await stopMeasurementEcg();
  }

  /// [stopMeasurementEcg] sends stop_ecg_packet to device.
  Future<void> stopMeasurementEcg() async {
    final control = points[Characteristics.DEVICE_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetStopEcg}');
    await control.write(Commands.packetStopEcg);
  }

  Future<bool> _checkValidation() async {
    final confirm = points[Characteristics.DEVICE_CONTROL_CONFIRM];
    if (confirm == null) return false;

    final confirmValue = await confirm.read();
    if (confirmValue.isEmpty) return false;
    if (confirmValue[0] == 2) return false;
    return true;
  }

  Future<void> _sendCommandReadEcg() async {
    final control = points[Characteristics.DEVICE_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetReadEcg}');
    await control.write(Commands.packetReadEcg);
  }

  Stream<List<int>> _getStreamEcg() async* {
    final oneChannel = points[Characteristics.ONE_CHANNEL];
    if (oneChannel == null) return;
    await oneChannel.setNotifyValue(true);
    await for (final res in oneChannel.value) {
      yield res;
    }
  }

  Stream<List<int>> _getDecompressedStreamEcg() async* {
    final events = streams.rawStreamEcg;
    if (events == null) return;
    await for (final event in events) {
      yield lzma.decode(event);
    }
  }

  Stream<List<int>>? get _events {
    if (isCompressedEcg) return streams.decompressedStreamEcg;
    return streams.rawStreamEcg;
  }

  Stream<List<double>> _getRecentStreamEcg() async* {
    final events = _events;
    if (events == null) return;
    final res = <double>[];
    await for (final event in events) {
      if (event.isEmpty) continue;
      final itemCount = event.length ~/ 3;
      res.clear();
      for (var i = 0; i < itemCount; i++) {
        final raw = toDouble(
          event[i * 3 + 0],
          event[i * 3 + 1],
          event[i * 3 + 2],
        );
        res.add(raw);
        // res.add(isFilteredEcg ? _filter(raw) : raw);
      }
      yield res;
    }
  }

  Stream<LeadState> _getLeadStateStream() async* {
    final leadState = points[Characteristics.LEAD_STATE];
    if (leadState == null) return;
    await leadState.setNotifyValue(true);
    await for (final res in leadState.value) {
      if (res.length < 4) continue;
      yield LeadState(
        ra: res[3] & 0x01 == 0x01,
        rl: true,
        la: res[3] >> 1 & 0x01 == 0x01,
        ll: true,
        v1: true,
        v2: true,
        v3: true,
        v4: true,
        v5: true,
        v6: true,
      );
    }
  }

  Stream<int> _getHeartRateStream() async* {
    final events = streams.recentStreamEcg;
    if (events == null) return;

    await for (final event in events) {
      if (event.isEmpty) continue;

      for (var i = 0; i < event.length; i++) {
        final output = _adaptiveHR.calculateHR(
          event[i],
          true,
        );

        if (output > 0) yield output;
      }
    }
  }

  Stream<List<double>> _getBufferedStreamEcg() async* {
    final events = streams.recentStreamEcg;
    if (events == null) return;
    final res = <double>[];
    await for (final event in events) {
      res.addAll(event);
      if (res.length > maxBufferStreamEcg) {
        res.removeRange(0, res.length - lengthBufferStreamEcg);
      }
      yield res;
    }
  }

  Stream<List<FlSpot>> _getBufferedPlotStreamEcg() async* {
    final events = streams.recentStreamEcg;
    if (events == null) return;
    final res = <FlSpot>[];

    final f = isFilteredEcg
        ? (double value) {
            return _dataFilter.applyFilter(
              0,
              value,
              true,
            );
          }
        : (double value) {
            return value;
          };

    var x = 0.0;
    await for (final event in events) {
      for (final i in Iterable<int>.generate(event.length)) {
        res.add(FlSpot(x + i, f(event[i])));
      }
      if (res.length > maxBufferStreamEcg) {
        res.removeRange(0, res.length - lengthBufferStreamEcg);
      }
      yield res;
      x += event.length;
    }
  }
}
