// ignore_for_file: public_member_api_docs,
// ignore_for_file: use_late_for_private_fields_and_variables

import 'dart:developer';

import 'package:fl_chart/fl_chart.dart';
import 'package:healthrian_ble_support_for_wearcardiot/healthrian_ble_support_for_wearcardiot.dart';
import 'package:healthrian_common_support/healthrian_common_support.dart';
import 'package:lzma/lzma.dart';

mixin Oxi {
  /// [maxBufferStreamOxi] decides a maximum length of buffered Stream.
  ///
  /// For 40 seconds buffer and 100 sampling rate,
  /// It uses 40 * 1000.
  int maxBufferStreamOxi = 40 * 100;

  /// [lengthBufferStreamOxi] decides a length of buffered Stream,
  /// after flushing buffer when reaching a maximum size.
  ///
  /// For 30 seconds buffer and 100 sampling rate,
  /// It uses 30 * 100.
  int lengthBufferStreamOxi = 30 * 100;

  /// [isFilteredOxi] decides whether pulse oxi signal is filtered.
  bool isFilteredOxi = true;

  /// [isCompressedOxi] decides whether pulse oxi signal is compressed.
  ///
  /// If it is `true`, pulse oxi signal is decompressed by LZMA algorithm.
  bool isCompressedOxi = false;

  int _itemCountOxiChannel = 2;
  int _itemCountOxiSample = 10;

  int get _totalCount => _itemCountOxiChannel * _itemCountOxiSample;

  bool isPaused = false;
}

extension BLEOxi on BLEManager {
  // TODO(chan150): 구현 필요 in common library
  double _filter(double value) => value;

  /// [startMeasurementOxi] sends read_oxi_packet to device,
  ///  and then, it registers all ecg streams of [streams].
  Future<void> startMeasurementOxi() async {
    await _sendCommandReadOxi();

    await _getConfigurationOxi();

    final isRunnable = await _checkValidation();
    print('isRunnable : $isRunnable');
    if (!isRunnable) return;

    streams
      ..rawStreamOxi = _getStreamOxi().broadcast
      ..decompressedStreamOxi = _getDecompressedStreamOxi().broadcast
      ..recentStreamOxi = _getRecentStreamOxi().broadcast
      ..bufferedStreamOxi = _getBufferedStreamOxi().broadcast
      ..bufferedPlotStreamOxi = _getBufferedPlotStreamOxi().broadcast;

    isPaused = false;
  }

  /// [pauseMeasurementOxi] pause measurements for Pulse Oxi.
  Future<void> pauseMeasurementOxi() async {
    if (isPaused) {
      await startMeasurementOxi();
      return;
    }
    isPaused = true;
    await stopMeasurementOxi();
  }

  /// [stopMeasurementOxi] sends stop_oxi_packet to device.
  Future<void> stopMeasurementOxi() async {
    final control = points[Characteristics.DEVICE_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetStopOxi}');
    await control.write(Commands.packetStopOxi);
  }

  Future<bool> _checkValidation() async {
    final confirm = points[Characteristics.DEVICE_CONTROL_CONFIRM];
    if (confirm == null) return false;

    final confirmValue = await confirm.read();
    if (confirmValue.isEmpty) return false;
    if (confirmValue[0] == 2) return false;
    return true;
  }

  Future<void> _sendCommandReadOxi() async {
    final control = points[Characteristics.DEVICE_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetReadOxi}');
    await control.write(Commands.packetReadOxi);
  }

  Future<void> _getConfigurationOxi() async {
    final configuration = points[Characteristics.PULSE_OXIMETER_CONFIGURATION];
    if (configuration == null) return;

    final configurationValue = await configuration.read();

    if (configurationValue.length < 2) return;

    _itemCountOxiChannel = configurationValue[0];
    _itemCountOxiSample = configurationValue[1];
  }

  Stream<List<int>> _getStreamOxi() async* {
    final continuous = points[Characteristics.PULSE_OXIMETER_CONTINUOUS];
    if (continuous == null) return;
    await continuous.setNotifyValue(true);
    await for (final res in continuous.value) {
      yield res;
    }
  }

  Stream<List<int>> _getDecompressedStreamOxi() async* {
    final events = streams.rawStreamOxi;
    if (events == null) return;
    await for (final event in events) {
      yield isCompressedOxi ? lzma.decode(event) : event;
    }
  }

  Stream<List<double>> _getRecentStreamOxi() async* {
    final events = streams.rawStreamOxi;
    if (events == null) return;
    final res = <double>[];
    await for (final event in events) {
      if (event.isEmpty) continue;
      res.clear();
      for (var i = 0; i < _totalCount; i++) {
        final raw = toDouble(
          event[i * 3 + 0],
          event[i * 3 + 1],
          event[i * 3 + 2],
        );
        res.add(isFilteredOxi ? _filter(raw) : raw);
      }
      yield res;
    }
  }

  Stream<List<List<double>>> _getBufferedStreamOxi() async* {
    final events = streams.recentStreamOxi;
    if (events == null) return;

    final res = List<List<double>>.generate(
      _itemCountOxiChannel,
      (index) => <double>[],
    );

    await for (final event in events) {
      for (var index = 0; index < _totalCount; index++) {
        res[index % _itemCountOxiChannel].add(event[index]);
      }

      for (final r in res) {
        if (r.length > maxBufferStreamOxi) {
          r.removeRange(0, r.length - lengthBufferStreamOxi);
        }
      }
      yield res;
    }
  }

  Stream<List<List<FlSpot>>> _getBufferedPlotStreamOxi() async* {
    final events = streams.recentStreamOxi;
    if (events == null) return;

    var x = 0.0;
    final res = List<List<FlSpot>>.generate(
      _itemCountOxiChannel,
      (index) => <FlSpot>[],
    );

    await for (final event in events) {
      for (var index = 0; index < _totalCount; index++) {
        res[index % _itemCountOxiChannel].add(
          FlSpot(
            x + index ~/ _itemCountOxiChannel,
            event[index],
          ),
        );
      }

      for (final r in res) {
        if (r.length > maxBufferStreamOxi) {
          r.removeRange(0, r.length - lengthBufferStreamOxi);
        }
      }
      yield res;

      x += event.length ~/ _itemCountOxiChannel;
    }
  }
}
